<!doctype html>
<html lang="en">

<head>
  <meta charset="utf-8">
  <meta http-equiv="x-ua-compatible" content="ie=edge">
  <title>{{ appname }}</title>
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <link rel="stylesheet" href="style.scss">
</head>

<body>
  {%! set base = '.' %}
  {%! include template-navbar.html %}
  <div class="container my-5">
    <div class="jumbotron text-center">
      <h1 class="display-4">Welcome to Gramex</h1>
      <a class="btn btn-primary btn-lg my-5" href="https://gramener.com/gramex/guide/" role="button">Learn more</a>
    </div>
  </div>

  <script src="ui/bootstrap5/dist/js/bootstrap.bundle.min.js"></script>
</body>

</html>
