/*! Themestr.app `Bootstrap Purple` Bootstrap 4.3.1 theme */

$primary: #563d7c;
$secondary: #868e96;
$success: #28a745;
$danger: #dc3545;
$info: #17a2b8;
$warning: #ffc107;
$light: #f8f9fa;
$dark: #343434;
@import "gramexui";

// Add SASS theme customizations here..
