/*! Themestr.app `Greyson` Bootstrap 4.3.1 theme */
@import url(https://fonts.googleapis.com/css?family=Muli:200,300,400,700);
$font-family-base: Muli;
@import url(https://fonts.googleapis.com/css?family=Oswald:200,300,400,700);
$headings-font-family: Oswald;

$primary: #2f3c48;
$secondary: #6f7f8c;
$success: #3e4d59;
$danger: #cc330d;
$info: #5c8f94;
$warning: #6e9fa5;
$light: #eceeec;
$dark: #1e2b37;

/*! Import Bootstrap 4 variables */
@import "functions";
@import "variables";

$enable-rounded: false;
@import "gramexui";

// Add SASS theme customizations here..
