/*! Themestr.app `Lovey` Bootstrap 4.3.1 theme */
@import url(https://fonts.googleapis.com/css?family=Muli:200,300,400,700);
$font-family-base: Muli;
@import url(https://fonts.googleapis.com/css?family=Lato:200,300,400,700);
$headings-font-family: Lato;

$primary: #d80e70;
$secondary: #dddddd;
$success: #4ef037;
$danger: #fa4659;
$info: #00bbf0;
$warning: #f4f133;
$light: #eeeeee;
$dark: #353535;
@import "gramexui";

// Add SASS theme customizations here..
