/*! Themestr.app `Herbie` Bootstrap 4.3.1 theme */
@import url(https://fonts.googleapis.com/css?family=Nunito:200,300,400,700);
$font-family-base: Nunito;
@import url(https://fonts.googleapis.com/css?family=Crete+Round:200,300,400,700);
$headings-font-family: Crete Round;

$primary: #083358;
$secondary: #f67280;
$success: #0074e4;
$danger: #ff4057;
$info: #74dbef;
$warning: #fc3c3c;
$light: #f2f2f0;
$dark: #072247;
@import "gramexui";

// Add SASS theme customizations here..
