/*! Themestr.app `Blue Voltage` Bootstrap 4.3.1 theme */
@import url(https://fonts.googleapis.com/css?family=Nunito:200,300,400,700);
$headings-font-family: Nunito;

$primary: #1633ff;
$secondary: #a0b7bc;
$success: #1633ff;
$danger: #00f900;
$info: #7ebcfa;
$warning: #00f900;
$light: #eef0f2;
$dark: #000633;

/*! Import Bootstrap 4 variables */
@import "functions";
@import "variables";

$body-bg: #ffffff;
$display1-weight: 600;
$display2-weight: 600;
$display3-weight: 600;
$btn-border-radius: 1.3rem;
$btn-border-radius-lg: 1.6rem;
$btn-border-radius-sm: 0.8rem;
@import "gramexui";

// Add SASS theme customizations here..
