/*! Themestr.app `Hello Kiddie` Bootstrap 4.3.1 theme */
@import url(https://fonts.googleapis.com/css?family=Raleway:200,300,400,700);
$headings-font-family: Raleway;

$primary: #f6648c;
$secondary: #cacaca;
$success: #035668;
$danger: #ff444f;
$info: #3f81c7;
$warning: #0de2ea;
$light: #eaeaea;
$dark: #223322;
@import "gramexui";

// Add SASS theme customizations here..
html {
  font-size: 0.8rem;
}
@include media-breakpoint-up(sm) {
  html {
    font-size: 0.9rem;
  }
}
@include media-breakpoint-up(md) {
  html {
    font-size: 1rem;
  }
}
@include media-breakpoint-up(lg) {
  html {
    font-size: 1.1rem;
  }
}
