/*! Themestr.app `Poypull` Bootstrap 4.3.1 theme */
@import url(https://fonts.googleapis.com/css?family=Muli:200,300,400,700);
$headings-font-family: Muli;

$primary: #7f0fff;
$secondary: #bed6d5;
$success: #420084;
$danger: #f2460d;
$info: #7ebcfa;
$warning: #ff9933;
$light: #eef0f2;
$dark: #000633;

/*! Import Bootstrap 4 variables */
@import "functions";
@import "variables";

$display1-weight: 600;
$display2-weight: 600;
$display3-weight: 600;
$navbar-dark-color: #f3f3f3;
@import "gramexui";

// Add SASS theme customizations here..
