/*! Themestr.app `Hootstrap` Bootstrap 4.3.1 theme */
@import url(https://fonts.googleapis.com/css?family=Open+Sans:200,300,400,700);
$font-family-base: Open Sans;
@import url(https://fonts.googleapis.com/css?family=Autour+One:200,300,400,700);
$headings-font-family: Autour One;

$primary: #528078;
$secondary: #eed75a;
$success: #fec100;
$danger: #703b3b;
$info: #63e792;
$warning: #ffe869;
$light: #fdfbf7;
$dark: #555555;
@import "gramexui";

// Add SASS theme customizations here..
