/*! Themestr.app `Signal` Bootstrap 4.3.1 theme */
@import url(https://fonts.googleapis.com/css?family=Nunito:200,300,400,700);
$font-family-base: Nunito;
@import url(https://fonts.googleapis.com/css?family=Comfortaa:200,300,400,700);
$headings-font-family: Comfortaa;

$primary: #ff304f;
$secondary: #cacaca;
$success: #015668;
$danger: #06648c;
$info: #0f81c7;
$warning: #0de2ea;
$light: #ececec;
$dark: #222222;

/*! Import Bootstrap 4 variables */
@import "functions";
@import "variables";

$gray-300: #e0e0e0;
@import "gramexui";

// Add SASS theme customizations here..
