/*! Themestr.app `Tequila` Bootstrap 4.3.1 theme */
@import url(https://fonts.googleapis.com/css?family=Voltaire:200,300,400,700);
$headings-font-family: Voltaire;

$primary: #2f414a;
$secondary: #f47b53;
$success: #420084;
$danger: #f2460d;
$info: #7ebcfa;
$warning: #ff9933;
$light: #eef0f2;
$dark: #000633;
@import "gramexui";

// Add SASS theme customizations here..
