/*! Themestr.app `Fresca` Bootstrap 4.3.1 theme */
@import url(https://fonts.googleapis.com/css?family=Muli:200,300,400,700);
$headings-font-family: Muli;

$primary: #07689f;
$secondary: #c9d6df;
$success: #11d3bc;
$danger: #f67280;
$info: #a2d5f2;
$warning: #ff7e67;
$light: #fafafa;
$dark: #4e4e4e;
@import "gramexui";

// Add SASS theme customizations here..
